/*
  Copyright (C) 2004  Jonathan Harrington jonathan@skynet

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package sync4j.source;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.util.Properties;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.logging.Logger;
import java.util.logging.Level;
import java.security.Principal;
import java.sql.Timestamp;

import sync4j.framework.logging.Sync4jLogger;
import sync4j.framework.engine.SyncItemImpl;
import sync4j.framework.engine.SyncItem;
import sync4j.framework.engine.SyncItemKey;
import sync4j.framework.engine.SyncProperty;
import sync4j.framework.engine.SyncItemState;
import sync4j.framework.engine.source.SyncSource;
import sync4j.framework.engine.source.SyncSourceException;
import sync4j.framework.engine.source.AbstractSyncSource;
import sync4j.framework.tools.Base64;
import sync4j.framework.tools.IOTools;

import javax.mail.*;
import javax.mail.internet.*;
import java.io.IOException;
import java.io.NotSerializableException;

import javax.naming.InitialContext;
import javax.naming.Context;
import javax.naming.NamingException;
import javax.rmi.PortableRemoteObject;
import java.rmi.RemoteException;
import javax.ejb.CreateException;
import java.util.Properties;
import java.util.ArrayList;

 	/**
	*
	* @author Jonathan Harrington
	*
	*/
	public class GroupwareSyncSource extends SyncSourceBase implements SyncSource, Serializable {

	// --------------------------------------------------------------- Constants

	public static final String LOG_NAME = "sync4j.framework.engine";
	public static final String DATABASE_FILE_NAME = "sync.db";
	public static final String DATABASE_HEADER = "Groupware file database";


	// -------------------------------------------------------------- Properties

	/**
	* The directory where database files are stored (the default is the current
	* directory) - read/write
	*/
	private String sourceDirectory = ".";

	public void setSourceDirectory(String sourceDirectory) {
		this.sourceDirectory = sourceDirectory;
	}

	public String getSourceDirectory() {
		return this.sourceDirectory;
	}


	private String jdbcDriver = "org.hsqldb.jdbcDriver";
	public void setJdbcDriver(String jdbcDriver) {
		this.jdbcDriver = jdbcDriver;
	}
	public String getJdbcDriver() {
		return this.jdbcDriver;
	}

	private String jdbcUrl = "jdbc:hsqldb:hsql://localhost:1476";
	public void setJdbcUrl(String jdbcUrl) {
		this.jdbcUrl = jdbcUrl;
	}
	public String getJdbcUrl() {
		return this.jdbcUrl;
	}

	// ------------------------------------------------------------ Constructors

	public GroupwareSyncSource() {}

	public GroupwareSyncSource(String name) {
		super(name);
		setSourceURI(name);
	}

	public GroupwareSyncSource(String name, String jdbcDriver, String jdbcUrl) {
		super(name);
		setSourceURI(name);
		this.jdbcDriver = jdbcDriver;
		this.jdbcUrl = jdbcUrl;
	}

	// ---------------------------------------------------------- Public Methods

	/*
	* @see SyncSource
	*/
	public SyncItem setSyncItem(Principal principal, SyncItem syncItem) throws SyncSourceException
	{
		String fileName = syncItem.getKey().getKeyAsString();
		byte[] fileContent = (byte[])syncItem.getPropertyValue(SyncItem.PROPERTY_BINARY_CONTENT);

		if(fileContent == null) { fileContent = new byte[0]; }


		// Check if item is in the sync database already
		SyncItem isItThere = getSyncItemFromId(principal, syncItem.getKey());
		if(isItThere == null)
  		{
			// Add file to datastore
			this.addFile(principal, getSourceURI(), fileContent, fileName);
		} else {
			// Check if new item is different from server item
			// and if so, delete old one and upload new item
			byte[] serverFileContent = (byte[])isItThere.getPropertyValue(SyncItem.PROPERTY_BINARY_CONTENT);
			if(serverFileContent != fileContent)
			{
				this.deleteFile(principal, getSourceURI(), fileName);
				this.addFile(principal, getSourceURI(), fileContent, fileName);
			}

		}

		Timestamp t = (Timestamp)syncItem.getPropertyValue(SyncItem.PROPERTY_TIMESTAMP);

		// Update Database
		this.setState(principal, fileName, SyncItemState.UPDATED, t);

		SyncItem newSyncItem = new SyncItemImpl(this, fileName, SyncItemState.NEW);
		newSyncItem.setProperties(syncItem.getProperties());

		return newSyncItem;
	}

	/**
	* @see SyncSource
	*/
	public SyncItem[] getAllSyncItems(Principal principal) throws SyncSourceException
	{
		return filterSyncItems(principal, null, "UNS");
	}

	/*
	* @see SyncSource
	*/
	public SyncItem[] getNewSyncItems(Principal principal, Timestamp since) throws SyncSourceException
	{
		return filterSyncItems(principal, since, String.valueOf(SyncItemState.NEW));
	}

	/*
	* @see SyncSource
	*/
	public SyncItem[] getDeletedSyncItems(Principal principal, Timestamp since) throws SyncSourceException
	{
		return filterSyncItems(principal, since, String.valueOf(SyncItemState.DELETED));
	}

	/**
	* @see SyncSource
	*/
	public SyncItem[] getUpdatedSyncItems(Principal principal, Timestamp since) throws SyncSourceException
	{
		return filterSyncItems(principal, since, String.valueOf(SyncItemState.UPDATED));
	}

	/*
	* @see SyncSource
	*/
	public void removeSyncItem(Principal principal, SyncItem syncItem) throws SyncSourceException
	{
		String fileName = syncItem.getKey().getKeyAsString();
		this.deleteFile(principal, getSourceURI(), fileName);

		Timestamp t = (Timestamp)syncItem.getPropertyValue(SyncItem.PROPERTY_TIMESTAMP);
		setState(principal, fileName, SyncItemState.DELETED, t);

	}

	/*
	* @see SyncSource
	*/
	public void removeSyncItems(Principal principal, SyncItem[] syncItems) throws SyncSourceException
	{
		ArrayList syncItemsInError = new ArrayList();

		for(int i=0; i<syncItems.length; i++)
		{
			try{
				removeSyncItem(principal, syncItems[i]);
			} catch (SyncSourceException e) {
				syncItemsInError.add(syncItems[i]);
			}
		}

		if (syncItemsInError.size() > 0)
		{
			throw new SyncSourceException("Error deleting the following items: "+ syncItemsInError.toString());
		}
	}

	/*
	* @see SyncSource
	*/
	public SyncItem getSyncItemFromId(Principal principal, SyncItemKey syncItemKey) throws SyncSourceException
	{
		SyncItem[] syncItems = getSyncItemsFromIds(principal, new SyncItemKey[] {syncItemKey});

		if((syncItems == null) || (syncItems.length == 0))
		{
			return null;
		}
		return syncItems[0];
	}

	/*
	* @see SyncSource
	*/
	public SyncItem[] getSyncItemsFromIds(Principal principal, SyncItemKey[] syncItemKeys) throws SyncSourceException
	{
		Properties syncDatabase = updateSyncDatabase(principal);

		ArrayList syncItems = new ArrayList();

		SyncItem syncItem = null;
		String fileName = null, stateString = null, file = null;
		char fileState = ' ';

		for(int i=0; ((syncItemKeys != null) && (i<syncItemKeys.length)); i++)
		{
			fileName = syncItemKeys[i].getKeyAsString();
			stateString = syncDatabase.getProperty(fileName);

			if(stateString != null)
			{
				fileState = stateFromStateString(stateString);
				syncItem = new SyncItemImpl(this, fileName, fileState);

				file = this.getFile(principal, getSourceURI(), fileName);
				syncItem.setProperty(new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT, file.getBytes()));

				syncItems.add(syncItem);
			}
		}

		return (SyncItem[])syncItems.toArray(new SyncItem[syncItems.size()]);
	}

	/**
	* The same as <i>getSyncItemFromId()</i>
	* @see SyncSource
	*
	*/
	public SyncItem getSyncItemFromTwin(Principal principal, SyncItem syncItem) throws SyncSourceException
	{
		return getSyncItemFromId(principal, syncItem.getKey());
	}

	/**
	* The same as <i>getSyncItemFromIds()</i>
	*
	* @see SyncSource
	*/
	public SyncItem[] getSyncItemsFromTwins(Principal principal, SyncItem[] syncItems) throws SyncSourceException
	{
		ArrayList items = new ArrayList();

		for(int i=0; i<syncItems.length; i++)
		{
			items.add(getSyncItemFromId(principal, syncItems[i].getKey()));
		}

		return (SyncItem[])items.toArray(new SyncItem[items.size()]);
	}

	/*
	* @see SyncSource
	*/
	public SyncItem[] setSyncItems(Principal principal, SyncItem[] syncItems) throws SyncSourceException
	{
		ArrayList syncItemsInError = new ArrayList();
		ArrayList ret		   = new ArrayList();

		for(int i=0; i<syncItems.length; i++)
		{
			try {
				ret.add(setSyncItem(principal, syncItems[i]));
			} catch (SyncSourceException e)
			{
				syncItemsInError.add(syncItems[i]);
			}
		}

		if(syncItemsInError.size() > 0)
		{
			throw new SyncSourceException("Error setting the following items: "+ syncItemsInError.toString());
		}

		return (SyncItem[])ret.toArray(new SyncItem[] {});
	}

	// ---------------------------------------------------------- private Methods
	private Properties updateSyncDatabase(Principal principal)
	{
		Properties syncDatabase = new Properties();
		try{
			File fileSyncDatabase = getDatabaseFile(principal);

			if(fileSyncDatabase.exists())
			{
				FileInputStream fis = new FileInputStream(fileSyncDatabase);
				syncDatabase.load(fis);
				fis.close();
			}


			String[] fileNames = null;
			String[] files = null;

			files = this.getFiles(principal, getSourceURI());

			fileNames = this.getFileNames(principal, getSourceURI());

			// Get a list of files on server
			ArrayList existingFiles = new ArrayList();

			// FIXME URGENCY medium has to be a better way.
			for(int i=0; i<fileNames.length; i++)
			{
				existingFiles.add(fileNames[i]);
			}

			// Get files in the database
			Enumeration databaseFiles = syncDatabase.propertyNames();

			String state = null;
			String fileName = null;
			long lastModified = 0;

			for(Iterator i = existingFiles.iterator(); i.hasNext();)
			{
				fileName = (String)i.next();
				lastModified = getLastModified(principal, getSourceURI(), fileName);
				state = syncDatabase.getProperty(fileName);

				if(state != null)
				{
					//The file is already in database
					if(lastModified > lastModifiedFromStateString(state))
					{
						state = buildStateString(SyncItemState.UPDATED, lastModified);
						syncDatabase.setProperty(fileName, state);
					}
				} else {
					// The file is not in the Database
					state = buildStateString(SyncItemState.NEW, lastModified);
					syncDatabase.setProperty(fileName, state);
				}
			} // next i

			for(; databaseFiles.hasMoreElements();)
			{
				fileName = (String)databaseFiles.nextElement();

				if(!existingFiles.contains(fileName))
				{
					state = buildStateString(SyncItemState.DELETED, System.currentTimeMillis());
					syncDatabase.setProperty(fileName, state);
				}
			} // next

			// Save and return
			FileOutputStream fos = new FileOutputStream(fileSyncDatabase);
			syncDatabase.store(fos, DATABASE_HEADER);
			fos.close();
		} catch (SyncSourceException e) {
			Logger log = Sync4jLogger.getLogger("source");
			if(log.isLoggable(Level.SEVERE))
			{
				log.severe("Error updating sync database: "+e.getMessage());
			}
			Logger.getLogger(LOG_NAME).throwing(getClass().getName(), "updateSyncDatabase", e);
		} catch (IOException e) {
			Logger log = Sync4jLogger.getLogger("source");
			if(log.isLoggable(Level.SEVERE))
			{
				log.severe("Error updating sync database: "+e.getMessage());
			}
			Logger.getLogger(LOG_NAME).throwing(getClass().getName(), "updateSyncDatabase", e);
		}

		return syncDatabase;
	}

	private SyncItem[] filterSyncItems(Principal principal, Timestamp since, String states) throws SyncSourceException
	{
		Properties syncDatabase = updateSyncDatabase(principal);
		ArrayList syncItems = new ArrayList();

		long fileTimestamp,
		sinceTimestamp = (since ==null) ? -1 : since.getTime();

		SyncItem syncItem = null;
		String fileName = null;
		String stateString = null;
		char fileState;

		String file = null;
		String[] files = null;
		String[] fileNames = null;

		files = this.getFiles(principal, getSourceURI());
		fileNames = this.getFileNames(principal, getSourceURI());

		for(Enumeration e = syncDatabase.keys(); e.hasMoreElements();)
		{
			fileName = (String)e.nextElement();
			stateString = (String)syncDatabase.get(fileName);
			fileState = stateFromStateString(stateString);
			file = this.getFile(principal, getSourceURI(), fileName);

			if(states.indexOf(fileState) >= 0)
			{
				fileTimestamp = lastModifiedFromStateString(stateString);

				if((fileTimestamp - sinceTimestamp) > 1000)
				{
					syncItem = new SyncItemImpl(this, fileName, fileState);

					if (fileState != SyncItemState.DELETED)
					{
						syncItem.setProperty(new SyncProperty(SyncItem.PROPERTY_BINARY_CONTENT, file.getBytes()));
					}
					syncItems.add(syncItem);
				}
			}
		} // Next e
		return (SyncItem[])syncItems.toArray(new SyncItem[0]);
	}


	// ------------------------------------------ Update file db methods
	private void setState(Principal principal, String file, char state, Timestamp t) throws SyncSourceException
	{
		try {
			Properties syncDatabase = new Properties();

			File fileSyncDatabase = getDatabaseFile(principal);

			//
			// Reads the existing database
			//
			if (fileSyncDatabase.exists()) {
				FileInputStream fis = new FileInputStream(fileSyncDatabase);
				syncDatabase.load(fis);
				fis.close();
			}

			syncDatabase.setProperty(file, buildStateString(state, t.getTime()));

			FileOutputStream fos = new FileOutputStream(fileSyncDatabase);
			syncDatabase.store(fos, DATABASE_HEADER);
			fos.close();
		} catch (IOException e) {
			Logger.getLogger(LOG_NAME).throwing(getClass().getName(), "setState", e);
		}
	}

	private File getDatabaseFile(Principal principal)
	{
		try {
			// Directory is sourceDirectory which is set in Syncsource bean descriptor
			// followed by the principal. (username.device)
			String dir = sourceDirectory+File.separator+principal.toString();
			// Db filename is principal + source URI + sync.db.
			// source URI is passed by client e.g gwcontact
			String dbName = principal.toString()+"."+getSourceURI()+"."+DATABASE_FILE_NAME;

			// If directory doesnt exist, create it.
			if (!new File(dir).exists())
			{
				if(!new File(dir).mkdirs()) {
					throw new java.io.IOException("Unable to create directory: "+dir);
				}
			}

			return new File(dir, dbName);
		} catch (IOException e) {
			Logger.getLogger(LOG_NAME).throwing(getClass().getName(), "getDatabaseFile", e);
		}
			return new File(sourceDirectory, DATABASE_FILE_NAME);
	}

	private String buildStateString(char state, long lastModified)
	{
		return state+String.valueOf(lastModified);
	}

	private long lastModifiedFromStateString(String state)
	{
		return Long.parseLong(state.substring(1));
	}

	private char stateFromStateString(String state)
	{
		if ((state == null) || (state.length() == 0)) { return SyncItemState.UNKNOWN; }

		return state.charAt(0);
	}

	// ------------------------------------------ Not used by Sync4j yet
	public SyncItemKey[] getNewSyncItemKeys(Principal principal, Timestamp since) throws SyncSourceException { return null;}
	public SyncItemKey[] getDeletedSyncItemKeys(Principal principal, Timestamp since) throws SyncSourceException { return null; }
	public SyncItemKey[] getUpdatedSyncItemKeys(Principal principal,Timestamp since) throws SyncSourceException { return null; }
}
